@@ -3,7 +3,7 @@
    Parser for dhclient config and lease files... */
 
 /*
- * Copyright (c) 2004-2007 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2007,2009 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -22,12 +22,12 @@
  *   950 Charter Street
  *   Redwood City, CA 94063
  *   <info@isc.org>
- *   http://www.isc.org/
+ *   https://www.isc.org/
  *
  * This software has been written for Internet Systems Consortium
  * by Ted Lemon in cooperation with Vixie Enterprises and Nominum, Inc.
  * To learn more about Internet Systems Consortium, see
- * ``http://www.isc.org/''.  To learn more about Vixie Enterprises,
+ * ``https://www.isc.org/''.  To learn more about Vixie Enterprises,
  * see ``http://www.vix.com''.   To learn more about Nominum, Inc., see
  * ``http://www.nominum.com''.
  */
@@ -499,6 +499,7 @@ void parse_client_statement (cfile, ip, config)
 			return;
 		}
 
+		known = 0;
 		status = parse_option_name(cfile, 1, &known, &option);
 		if (status != ISC_R_SUCCESS || option == NULL)
 			return;
@@ -518,6 +519,18 @@ void parse_client_statement (cfile, ip, config)
 			option_dereference(&option, MDL);
 			return;
 		}
+
+		/*
+		 * If the option was known, remove it from the code and name
+		 * hash tables before redefining it.
+		 */
+		if (known) {
+			option_name_hash_delete(option->universe->name_hash,
+						option->name, 0, MDL);
+			option_code_hash_delete(option->universe->code_hash,
+						&option->code, 0, MDL);
+		}
+
 		parse_option_code_definition(cfile, option);
 		option_dereference(&option, MDL);
 		return;
