@@ -3,7 +3,7 @@
    Subroutines for dealing with connections. */
 
 /*
- * Copyright (c) 2004,2007 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004,2007,2009 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1999-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -22,12 +22,12 @@
  *   950 Charter Street
  *   Redwood City, CA 94063
  *   <info@isc.org>
- *   http://www.isc.org/
+ *   https://www.isc.org/
  *
  * This software has been written for Internet Systems Consortium
  * by Ted Lemon in cooperation with Vixie Enterprises and Nominum, Inc.
  * To learn more about Internet Systems Consortium, see
- * ``http://www.isc.org/''.  To learn more about Vixie Enterprises,
+ * ``https://www.isc.org/''.  To learn more about Vixie Enterprises,
  * see ``http://www.vix.com''.   To learn more about Nominum, Inc., see
  * ``http://www.nominum.com''.
  */
@@ -674,16 +674,16 @@ static isc_result_t omapi_connection_connect_internal (omapi_object_t *h)
 			 (struct sockaddr *)&c -> local_addr, &sl) < 0) {
 	}
 
-	/* Disconnect from I/O object, if any. */
-	if (h -> outer)
-		omapi_unregister_io_object (h);
+	/* Reregister with the I/O object.  If we don't already have an
+	   I/O object this turns into a register call, otherwise we simply
+	   modify the pointers in the I/O object. */
 
-	status = omapi_register_io_object (h,
-					   omapi_connection_readfd,
-					   omapi_connection_writefd,
-					   omapi_connection_reader,
-					   omapi_connection_writer,
-					   omapi_connection_reaper);
+	status = omapi_reregister_io_object (h,
+					     omapi_connection_readfd,
+					     omapi_connection_writefd,
+					     omapi_connection_reader,
+					     omapi_connection_writer,
+					     omapi_connection_reaper);
 
 	if (status != ISC_R_SUCCESS) {
 		omapi_disconnect (h, 1);
